//Autor: Ate Akaydn
//Date: 12.05.2008
/*
	This is the voronoi generator class. It basically generates the voronoi diagram from a list of given sites.
	Sites can be points, lines, curves, polygons, sets...
	Voronoi generator generates OpenGL display lists of cones and tents(and other primitives for complex sites) for
	Proper rasterization.
	These primitives are generated with respect to the size of the viewport and the error threshold
*/

#ifndef VG_VORGEN2D_H
#define VG_VORGEN2D_H

#include "vg_site2d.h"
#include "vg_point2d.h"
#include "vg_line2d.h"
#include "vg_polygon2d.h"
#include "vg_constants.h"
#include "vg_decoder.h"

#include <iostream>
#include <fstream>
#include <vector>

class VG_Vorgen2d{


private:

	//Size of the voronoi region
	double size_x, size_y;

	//Maximum radius
	double r;

	//Error threshold
	double er_thres;

	//Decoder
	VG_Decoder * decoder;

	//Vector of sites to be rastered
	std::vector<VG_Site2d*> sites;

	//These are the display lists which store standard primitives( cones, tents and others ) for rasterization
	GLuint * d_lists;

	//This routine generates a display list for a cone primitive for the given size and error threshold
	//r: is the estimated radius of the cone - should be
	void cone( ){

		assert( er_thres > 0 );

		//Step is the unit angle in between each triangle
		GLfloat stp =  2 * acos( (r-er_thres)/r );

		//Generate a new list
		d_lists[0] = glGenLists(1);


		glNewList( d_lists[0], GL_COMPILE );

		//Draw the cone mesh to the list
		glBegin( GL_TRIANGLE_FAN );

			glVertex3f( 0.0 , 0.0 ,0.0 );

			for( GLfloat angle = 0; angle < 360.0; angle+=stp ){

				glVertex3f( r*cos(angle*M_PI/180), r*sin(angle*M_PI/180), -r );

			}

			glVertex3f( r, 0, -r);

		glEnd();

        if( VG_DRAW_SITES ){
            glBegin( GL_POINTS );

                glColor3f( 0.0, 0.0, 0.0 );
                glVertex3f( 0.0, 0.0, 0.0 );

            glEnd();
        }

            glEndList();

	}

	//This routine generates a display list for a tent primitive for the given size and error threshold
	//r: is the estimated radius of the tent - should be
	void tent( ){


		//Generate a new list
		d_lists[1] = glGenLists(1);

		glNewList( d_lists[1], GL_COMPILE );

		//Draw the first quad
		glBegin( GL_QUAD_STRIP );
			glVertex3f( 0.0 ,GLfloat(r), -GLfloat(r) );
			glVertex3f( 1.0 ,GLfloat(r), -GLfloat(r) );
			glVertex3f( 0.0 ,0.0 ,0.0 );
			glVertex3f( 1.0 ,0.0 ,0.0 );
			glVertex3f( 0.0 ,-GLfloat(r), -GLfloat(r) );
			glVertex3f( 1.0 ,-GLfloat(r), -GLfloat(r) );
		glEnd();

        if( VG_DRAW_SITES ){

            glBegin( GL_LINES );

			glColor3f( 0.0, 0.0, 0.0 );
			glVertex3f( 0.0, 0.0, 0.0 );
			glVertex3f( 1.0, 0.0, 0.0 );

            glEnd();

        }

		glEndList();

	}

public:

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a square region, threshold and decoder size
	VG_Vorgen2d( double size, double thres, unsigned int dec_size ): size_x(size), size_y(size), er_thres(thres){

		//Clear the sites first
		sites.clear();

		decoder = new VG_Decoder( dec_size );

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a square region, threshold and takes an available decoder directly
	VG_Vorgen2d( double size, double thres, VG_Decoder * dc ): size_x(size), size_y(size), er_thres(thres){

		//Clear the sites first
		sites.clear();
		assert(dc!=NULL);
		decoder = dc;

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies the region, error threshold and decoder size
	VG_Vorgen2d( double sizex, double sizey, double thres, unsigned int dec_size ): size_x(sizex), size_y(sizey), er_thres(thres){

		//Clear the sites first
		sites.clear();

		decoder = new VG_Decoder( dec_size );

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies the region, error threshold and takes an available decoder directly
	VG_Vorgen2d( double sizex, double sizey, double thres, VG_Decoder * dc ): size_x(sizex), size_y(sizey), er_thres(thres){

		//Clear the sites first
		sites.clear();
		assert(dc!=NULL);
		decoder = dc;

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a square region
	VG_Vorgen2d( std::vector<VG_Site2d*> sts, double size, double thres, unsigned int dec_size ): size_x(size), size_y(size), er_thres(thres){

		//Clear the sites first
		sites.clear();

		//Fill the sites list
		int sid = 0;
		std::vector<VG_Site2d*>::iterator i = sts.begin();
		while( i!=sts.end() ){
			(*i)->setSiteId( sid );
			sites.push_back(*i);
			i++;
			sid++;
		}

		decoder = new VG_Decoder( dec_size );

		//std::cout << "test" << std::endl;

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a square region, error_threshold, takes an available decoder
	VG_Vorgen2d( std::vector<VG_Site2d*> sts, double size, double thres, VG_Decoder * dc ): size_x(size), size_y(size), er_thres(thres){

		//Clear the sites first
		sites.clear();

		//Fill the sites list
		int sid = 0;
		std::vector<VG_Site2d*>::iterator i = sts.begin();
		while( i!=sts.end() ){
			(*i)->setSiteId( sid );
			sites.push_back(*i);
			i++;
			sid++;
		}

		assert(dc!=NULL);
		decoder = dc;

		//std::cout << "test" << std::endl;

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone();
		tent();

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a rectangular region
	VG_Vorgen2d( std::vector<VG_Site2d*> sts, double sz_x, double sz_y, double thres, unsigned int dec_size ): size_x(sz_x), size_y(sz_y), er_thres(thres){

		//Clear the sites first
		sites.clear();

		//Fill the sites list
		int sid = 0;
		std::vector<VG_Site2d*>::iterator i = sts.begin();
		while( i!=sts.end() ){
			(*i)->setSiteId( sid );
			sites.push_back(*i);
			i++;
			sid++;
		}

		decoder = new VG_Decoder( dec_size );

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone( );
		tent( );

	}

	//Default constructor specifies voronoi generation size(or viewport) and error threshold for
	//rasterization primitives(tents and cones)
	//This one specifies a rectangular region, error threshold, and an available decoder
	VG_Vorgen2d( std::vector<VG_Site2d*> sts, double sz_x, double sz_y, double thres, VG_Decoder * dc ): size_x(sz_x), size_y(sz_y), er_thres(thres){

		//Clear the sites first
		sites.clear();

		//Fill the sites list
		int sid = 0;
		std::vector<VG_Site2d*>::iterator i = sts.begin();
		while( i!=sts.end() ){
			(*i)->setSiteId( sid );
			sites.push_back(*i);
			i++;
			sid++;
		}
        assert(dc!=NULL);
		decoder = dc;

		//Radius is equal to the length of the diagonal
		r = pow( pow( size_x, 2.0 ) + pow( size_y, 2.0 ), 0.5 );

		d_lists = new GLuint[2];
		cone( );
		tent( );

	}

	//Deconstructor
	~VG_Vorgen2d( ){

		glDeleteLists( d_lists[0], 1 );
		glDeleteLists( d_lists[1], 1 );

		delete decoder;

		delete[] d_lists;
	}

	//Add site
	void addSite( VG_Site2d * site ){
		site->setSiteId( sites.size( ) );
		sites.push_back( site );
	}

	//Clear Sites
	void clearSites(){
		sites.clear();
	}

	//Clear Sites
	void deleteSites(){
        if( sites.empty() ) return;
		std::vector<VG_Site2d *>::iterator i = sites.begin();
		while( i != sites.end() ){
               delete (*i);
               i++;
        }
	}

	//Visualization of the voronoi diagram
	void draw(){

		//Now draw the sites
		std::vector< VG_Site2d * >::iterator i = sites.begin();

		while( i!=sites.end() ){

            //std::cout << (*i)->getSiteId( ) << std::endl;

			(*i)->draw( r, er_thres, decoder );
			i++;

		}

	}
	//Visualization of the voronoi diagram... optimized with display lists
	void draw_d(){

		//Now draw the sites
		std::vector< VG_Site2d * >::iterator i = sites.begin();

		while( i!=sites.end() ){

            //std::cout << (*i)->getSiteId( ) << std::endl;
			(*i)->raster( d_lists, decoder );
			i++;

		}

	}
	unsigned int getSizeX() const{
		return size_x;
	}
	unsigned int getSizeY() const{
		return size_y;
	}

	//Maximum radius
	double getMaxRadius()const{
		return r;
	}

	//Error threshold
	double getErrorThreshold() const {
		return er_thres;
	}

	//Get display lists
	GLuint * getDisplayLists( ) const {
		return d_lists;
	}
	//Get size
	unsigned int getNumSites(){
        return sites.size();
    }

    VG_Decoder* getDecoder() {
        return decoder;
    }

    //Outputs the site information to the given file stream
	void toFileStream(  std::ofstream & fstrm ){

        fstrm << sites.size( ) << std::endl;
		//Now draw the sites
		std::vector< VG_Site2d * >::iterator i = sites.begin();

		while( i!=sites.end() ){

			(*i)->toFileStream( fstrm );
			i++;

		}
    };

    //Outputs the site information to the given file stream
	void fromFileStream(  std::ifstream & fstrm ){

        sites.clear( );

        unsigned int num_sites = 0;

        fstrm >> num_sites;

        std::cout << "#Sites:" << num_sites <<  std::endl;

        char info[50];
        char tmpBuffer[50];
		//Now read the sites
		for( int i = 0; i < num_sites; i++ ){

            fstrm >> info;
            std::cout << "Site Type:" << info << std::endl;

            if( strcmp( info, "point2d" ) == 0 ){
                double x,y;
                fstrm >> x;
                fstrm >> y;
                //std::cout << "x:" << x << ",y:" << y << std::endl;

                addSite( new VG_Point2d( (GLfloat)(x),(GLfloat)(y),i ) );
            }
            else if( strcmp( info, "line2d" ) == 0 ){

                double x1,y1;
                double x2,y2;

                fstrm >> x1;
                fstrm >> y1;

                fstrm >> x2;
                fstrm >> y2;
                //std::cout << "x:" << x << ",y:" << y << std::endl;
                VG_Line2d * line = new VG_Line2d( new VG_Point2d( (GLfloat)(x1),(GLfloat)(y1),i ),new VG_Point2d( (GLfloat)(x2),(GLfloat)(y2),i ),i );
                addSite( line );

            }
            else if( strcmp( info, "polygon2d" ) == 0 ){

                unsigned int numVertex;
                fstrm >> numVertex;

                VG_Polygon2d * p = new VG_Polygon2d( i );
                for( int j = 0; j < numVertex; j++ ){

                    double x,y;
                    fstrm >> x;
                    fstrm >> y;

                    p->addPoint( new VG_Point2d( (GLfloat)(x),(GLfloat)(y),i ) );

                }

                addSite( p );

            }
            else{
                std::cout << "ERROR: Site Type Undefined !" << std::endl;
                return;
            }

		}

    };


};

#endif
